
import json

from typing import Union, Callable, Optional, Any

from pysquril.exc import ParseError
from pysquril.parser import (
    Key,
    ArraySpecific,
    ArraySpecificSingle,
    ArraySpecificMultiple,
    ArrayBroadcastSingle,
    ArrayBroadcastMultiple,
    SelectTerm,
    WhereTerm,
    OrderTerm,
    RangeTerm,
    SetTerm,
    Clause,
    UriQuery,
)
from pysquril.utils import audit_table

class SqlGenerator(object):

    """
    Generic class, used to implement SQL code generation.

    """

    db_init_sql = None
    json_array_sql = None
    cascade_on_drop = False

    def __init__(
        self,
        table_name: str,
        uri_query: str,
        data: Union[list, dict] = None,
        backup_cutoff: Optional[str] = None,
        array_agg: Optional[bool] = False,
        table_name_func: Optional[Callable] = None,
        audit: bool = False,
    ) -> None:
        self.table_name = table_name
        self.uri_query = uri_query
        self.data = data
        self.parsed_uri_query = UriQuery(table_name, uri_query)
        self.table_name_func = table_name_func
        self.operators = {
            'eq': '=',
            'gt': '>',
            'gte': '>=',
            'lt': '<',
            'lte': '<=',
            'neq': '!=',
            'like': 'like',
            'ilike': 'ilike',
            'not': 'not',
            'is': 'is',
            'in': 'in'
        }
        if not self.json_array_sql:
            msg = 'Extending the SqlGenerator requires setting the class level property: json_array_sql'
            raise Exception(msg)
        self.has_aggregate_func = False
        self.select_query = self.sql_select(backup_cutoff, array_agg)
        self.update_query = self.sql_update()
        self.delete_query = self.sql_delete()
        self.message = self.uri_message()
        self.alter_query = self.sql_alter(audit)

    # Classes that extend the SqlGenerator must implement the following methods
    # they are called by functions that are mapped over terms in clauses
    # for each term, an appropriate piece of SQL needs to be returned.
    # What is appropriate, depends on the backend.

    def _gen_sql_key_selection(self, term: SelectTerm, parsed: Key) -> str:
        """
        Generate SQL for selecting a Key element.

        Called by _term_to_sql_select when generating the select
        part of the SQL.

        """
        raise NotImplementedError

    def _gen_sql_array_selection(self, term: SelectTerm, parsed: ArraySpecific) -> str:
        """
        Generate SQL for selecting an ArraySpecific element.

        Called by _term_to_sql_select when generating the select
        part of the SQL.

        """
        raise NotImplementedError

    def _gen_sql_array_sub_selection(
        self,
        term: SelectTerm,
        parsed: Union[
            ArraySpecificSingle,
            ArraySpecificMultiple,
            ArrayBroadcastSingle,
            ArrayBroadcastMultiple,
        ],
    ) -> str:
        """
        Generate SQL for selecting inside arrays.

        Called by _term_to_sql_select when generating the select
        part of the SQL.

        """
        raise NotImplementedError

    def _gen_sql_col(self, term: Union[SelectTerm, WhereTerm, OrderTerm]) -> str:
        """
        Generate a column reference from a term,
        used in where and order clauses.

        """
        raise NotImplementedError

    def _gen_sql_update(self, term: Key) -> list:
        """
        Generate a list of update expressions, from a term
        using the data passed to the constructor.

        """
        raise NotImplementedError

    def _clause_map_terms(self, clause: Clause, map_func: Callable) -> list:
        # apply a function to all Terms in a clause
        out = []
        for term in clause.parsed:
            res = map_func(term)
            out.append(res)
        return out

    # methods for mapping functions over terms in different types of clauses

    def select_map(self, map_func: Callable) -> Optional[list]:
        return self._clause_map_terms(self.parsed_uri_query.select, map_func) \
            if self.parsed_uri_query.select else None

    def where_map(self, map_func: Callable) -> Optional[list]:
        return self._clause_map_terms(self.parsed_uri_query.where, map_func) \
            if self.parsed_uri_query.where else None

    def order_map(self, map_func: Callable) -> Optional[list]:
        return self._clause_map_terms(self.parsed_uri_query.order, map_func) \
            if self.parsed_uri_query.order else None

    def range_map(self, map_func: Callable) -> Optional[list]:
        return self._clause_map_terms(self.parsed_uri_query.range, map_func) \
            if self.parsed_uri_query.range else None

    def set_map(self, map_func: Callable) -> Optional[list]:
        return self._clause_map_terms(self.parsed_uri_query.set, map_func) \
            if self.parsed_uri_query.set else None

    def group_by_map(self, map_func: Callable) -> Optional[list]:
        return self._clause_map_terms(self.parsed_uri_query.group_by, map_func) \
            if self.parsed_uri_query.group_by else None

    # term handler functions
    # mapped over terms in a clause
    # generates SQL for each term
    # SQL is generated by calling other functions
    # which are implemented for specific SQL backend implementations

    def _term_to_sql_select(self, term: SelectTerm) -> str:
        rev = term.parsed.copy()
        rev.reverse()
        first_done = False
        for parsed in rev:
            if isinstance(parsed, Key):
                if not first_done:
                    selection = self._gen_sql_key_selection(term, parsed)
            elif isinstance(parsed, ArraySpecific):
                selection = self._gen_sql_array_selection(term, parsed)
            elif (
                isinstance(parsed, ArraySpecificSingle)
                or isinstance(parsed, ArraySpecificMultiple)
                or isinstance(parsed, ArrayBroadcastSingle)
                or isinstance(parsed, ArrayBroadcastMultiple)
            ):
                selection = self._gen_sql_array_sub_selection(term, parsed)
            else:
                raise Exception(f'Could not parse {term.original}')
            first_done = True
        return selection

    def _term_to_sql_where(self, term: WhereTerm) -> str:
        groups_start = ''.join(term.parsed[0].groups_start)
        groups_end = ''.join(term.parsed[0].groups_end)
        combinator = term.parsed[0].combinator if term.parsed[0].combinator else ''
        col = self._gen_sql_col(term)
        op = term.parsed[0].op
        val = term.parsed[0].val
        try:
            int(val)
            val = f"'{val}'" if op in ['eq', 'neq'] else val
        except ValueError:
            if val == 'null' or op == 'in':
                val = f'{val}'
            else:
                val = self._maybe_float(val)
        if op.endswith('.not'):
            op = op.replace('.', ' ')
        elif op.startswith('not.'):
            op = op.replace('.', ' ')
        elif op == 'in':
            val = val.replace('[', '')
            val = val.replace(']', '')
            values = val.split(',')
            new_values = []
            for v in values:
                new = "'%s'" % v
                new_values.append(new)
            joined = ','.join(new_values)
            val = "(%s)" % joined
        else:
            op = self.operators[op]
        if 'like' in op or 'ilike' in op:
            val = val.replace('*', '%')
        out = f'{groups_start} {combinator} {col} {op} {val} {groups_end}'
        return out

    def _term_to_sql_order(self, term: OrderTerm) -> str:
        selection = self._gen_sql_col(term)
        direction = term.parsed[0].direction
        return f'order by {selection} {direction}'

    def _term_to_sql_range(self, term: RangeTerm) -> str:
        return f'limit {term.parsed[0].end} offset {term.parsed[0].start}'

    def _term_to_sql_update(self, term: SelectTerm) -> list:
        return self._gen_sql_update(term)

    # mapper methods - used by public methods

    def _gen_sql_select_clause(self, backup_cutoff: Optional[str] = None) -> str:
        if not backup_cutoff:
            table_reference = self.table_name
        else:
            table_reference = self._gen_select_with_retention(backup_cutoff)
        out = self.select_map(self._term_to_sql_select)
        if not out:
            sql_select = f'select data from {table_reference}'
        else:
            joined = ",".join(out)
            sql_select = f"select {self.json_array_sql}({joined}) data from {table_reference}"
        return sql_select

    def _gen_sql_where_clause(self) -> str:
        out = self.where_map(self._term_to_sql_where)
        if not out:
            sql_where = ''
        else:
            joined = ' '.join(out)
            sql_where = f'where {joined}'
        return sql_where

    def _gen_sql_order_clause(self) -> str:
        out = self.order_map(self._term_to_sql_order)
        if not out:
            return ''
        else:
            return out[0]

    def _gen_sql_range_clause(self) -> str:
        out = self.range_map(self._term_to_sql_range)
        if not out:
            return ''
        else:
            return out[0]

    def _gen_sql_group_by_clause(self) -> str:
        out = self.group_by_map(self._term_to_sql_select)
        if not out:
            return ''
        else:
            cols = ",".join(out)
            return f"group by {cols}"

    def _gen_select_with_retention(self, backup_cutoff: str) -> str:
        raise NotImplementedError

    def _gen_array_agg(self, query: str) -> str:
        raise NotImplementedError

    def _maybe_float(self, val) -> Union[str, float]:
        raise NotImplementedError

    # public methods - called by constructor

    def sql_select(self, backup_cutoff: Optional[str] = None, array_agg: Optional[bool] = False) -> str:
        _select = self._gen_sql_select_clause(backup_cutoff)
        _where = self._gen_sql_where_clause()
        _order = self._gen_sql_order_clause()
        _range = self._gen_sql_range_clause()
        _group_by = self._gen_sql_group_by_clause()
        query = f'{_select} {_where} {_order} {_group_by} {_range}'
        if array_agg and not self.has_aggregate_func:
            return self._gen_array_agg(query)
        else:
            return query

    def sql_update(self) -> str:
        """
        Implementation notes:

        When calling set_map with the _gen_sql_update function
        is called on all set terms, resulting in a list of
        potentential update clauses.

        Sqlite offers json_patch which allows updating
        multiple keys in a JSON structure in one SQL
        statement, negating the need for multiple update
        clauses. The result of set_map is a list of duplicate
        clauses in the case where more than one set term is
        provided.

        Postgres, however, offers jsonb_set, which can
        only be called for one key within a JSON structure
        per SQL statement. The result of set_map is a list
        of unique update clauses.

        Generating valid update statements requires resolving
        the differences in the functionality offered by sqlite
        and postgres.

        The solution is to turn the output of the set_map function
        into a set, removing duplicate entries. This way the sqlite
        variant always returns only one update clause, while the
        postgres variant may return more.

        For postgres the return value from this function is one
        or more update statements, depending on how many JSON keys
        are being changed. For sqlite it is always only one statement.

        """
        out = self.set_map(self._term_to_sql_update)
        if not out:
            return ''
        else:
            _where = self._gen_sql_where_clause()
            _query = ""
            for expr in set(out):
                _query += f"update {self.table_name} {expr} {_where}; "
            return _query

    def sql_delete(self) -> str:
        _where = self._gen_sql_where_clause()
        if not _where:
            query = f"drop table {self.table_name}"
            if self.cascade_on_drop:
                query += " cascade"
        else:
            query = f"delete from {self.table_name} {_where}"
        return query

    def uri_message(self) -> str:
        return self.parsed_uri_query.message

    def sql_alter(self, audit: bool) -> str:
        """
        Change the name of a table, if it does not
        exist an error is raised.

        """
        alter = self.parsed_uri_query.alter
        if not alter:
            return ""
        else:
            term = alter.parsed[0]
            element = term.parsed[0]
            if audit:
                new_name = self.table_name_func(audit_table(element.val), no_schema=True)
            else:
                new_name = self.table_name_func(element.val, no_schema=True)
            sql = f"alter table {self.table_name} rename to {new_name}"
            return sql


class SqliteQueryGenerator(SqlGenerator):

    """Generate SQL for SQLite json1 backed tables, from a given UriQuery."""

    db_init_sql = None
    json_array_sql = 'json_array'

    # Helper functions - used by mappers

    def _maybe_apply_function(self, term: SelectTerm, selection: str) -> str:
        if not term.func:
            return selection
        elif term.func == 'count':
            if term.original in ['*', '1']:
                selection = '1'
        else:
            if term.func.endswith('_ts'):
                term.func = term.func.replace('_ts', '')
        self.has_aggregate_func = True
        return f"{term.func}({selection})"

    def _gen_sql_key_selection(self, term: SelectTerm, parsed: Key) -> str:
        return self._maybe_apply_function(term, f"json_extract(data, '$.{term.original}')")

    def _gen_sql_array_selection(self, term: SelectTerm, parsed: ArraySpecific) -> str:
        return self._maybe_apply_function(term, f"json_extract(data, '$.{term.original}')")

    def _gen_sql_array_sub_selection(
        self,
        term: SelectTerm,
        parsed: Union[
            ArraySpecificSingle,
            ArraySpecificMultiple,
            ArrayBroadcastSingle,
            ArrayBroadcastMultiple,
        ],
    ) -> str:
        if (
            isinstance(parsed, ArraySpecificSingle)
            or isinstance(parsed, ArraySpecificMultiple)
        ):
            fullkey = f"and fullkey = '$.{term.bare_term}[{parsed.idx}]'"
            vals = 'vals'
        else:
            fullkey = ''
            vals = 'json_group_array(vals)'
        temp = []
        for key in parsed.sub_selections:
            temp.append(f"json_extract(value, '$.{key}')")
        sub_selections = ','.join(temp)
        sub_selections = f'json_array({sub_selections})' if len(temp) > 1 else f'{sub_selections}'
        selection = f"""
                (case when json_extract(data, '$.{term.bare_term}') is not null then (
                    select {vals} from (
                        select
                            {sub_selections} as vals
                        from (
                            select key, value, fullkey, path
                            from {self.table_name}, json_tree({self.table_name}.data)
                            where path = '$.{term.bare_term}'
                            {fullkey}
                            )
                        )
                    )
                else null end)
            """
        return self._maybe_apply_function(term, selection)

    def _gen_sql_col(self, term: Union[SelectTerm, WhereTerm, OrderTerm]) -> str:
        if isinstance(term, WhereTerm) or isinstance(term, OrderTerm):
            select_term = term.parsed[0].select_term
        elif isinstance(term, SelectTerm):
            select_term = term
        if len(select_term.parsed) > 1:
            test_select_term = select_term.parsed[-1]
            if isinstance(test_select_term, ArraySpecific):
                target = select_term.original
            elif isinstance(test_select_term, ArraySpecificSingle):
                _key = select_term.bare_term
                _idx = select_term.parsed[-1].idx
                _col = select_term.parsed[-1].sub_selections[0]
                target = f'{_key}[{_idx}].{_col}'
            else:
                target = select_term.bare_term
        else:
            if not (
                isinstance(select_term.parsed[0], Key) or
                isinstance(select_term.parsed[0], ArraySpecific)
            ):
                raise ParseError(f'Unsupported where term: {term.original}')
            if isinstance(select_term.parsed[0], ArraySpecific):
                target = select_term.original
            else:
                target = select_term.parsed[0].element
        col = f"json_extract(data, '$.{target}')"
        if isinstance(term, WhereTerm) and term.parsed[0].op in ['eq', 'neq']:
            col = f"cast ({col} as text)"
        return col

    def _gen_sql_update(self, term: SetTerm) -> str:
        key = term.parsed[0].select_term.bare_term
        if not self.data or key not in self.data.keys():
            raise ParseError(f'Target key of update: {key} not found in payload')
        new = json.dumps(self.data).replace("'", "''")
        return f"set data = json_patch(data, '{new}')"

    def _gen_select_with_retention(self, backup_cutoff: str) -> str:
        return f"(select * from {self.table_name} where json_extract(data, '$.timestamp') >= '{backup_cutoff}')a"

    def _gen_array_agg(self, query: str) -> str:
        return f"select json_group_array(data) from ({query})"

    def _maybe_float(self, val: Any) -> Union[str, float]:
        try:
            float(val)
            if str(float(val)) == str(val):
                return val
            else:
                return f"'{val}'"
        except ValueError:
            return f"'{val}'"


class PostgresQueryGenerator(SqlGenerator):

    cascade_on_drop = True
    json_array_sql = 'jsonb_build_array'
    db_init_sql = [
        """
        create or replace function filter_array_elements(data jsonb, keys text[])
            returns jsonb as $$
            declare key text;
            declare element jsonb;
            declare filtered jsonb;
            declare out jsonb;
            declare val jsonb;
            begin
                create temporary table if not exists info(v jsonb) on commit drop;
                for element in select jsonb_array_elements(data) loop
                    for key in select unnest(keys) loop
                        if filtered is not null then
                            filtered := filtered || jsonb_extract_path(element, key);
                        else
                            filtered := jsonb_extract_path(element, key);
                        end if;
                    if filtered is null then
                        filtered := '[]'::jsonb;
                    end if;
                    end loop;
                insert into info values (filtered);
                filtered := null;
                end loop;
                out := '[]'::jsonb;
                for val in select * from info loop
                    out := out || jsonb_build_array(val);
                end loop;
                return out;
            end;
        $$ language plpgsql;
        """,
        """
        create or replace function unique_data()
        returns trigger as $$
            begin
                NEW.uniq := md5(NEW.data::text);
                return new;
            end;
        $$ language plpgsql;
        """
    ]

    def _maybe_apply_function(self, term: SelectTerm, selection: str) -> str:
        if not term.func:
            return selection
        elif term.func == 'count':
            if term.original in ['*', '1']:
                selection = '1'
        else:
            if term.func in ['avg', 'sum', 'min', 'max']:
                selection = f"({selection})::int"
            if term.func.endswith('_ts'):
                term.func = term.func.replace('_ts', '')
        self.has_aggregate_func = True
        return f"{term.func}({selection})"

    def _gen_select_target(self, term_attr: str) -> str:
        return term_attr.replace('.', ',') if '.' in term_attr else term_attr

    def _gen_sql_key_selection(self, term: SelectTerm, parsed: Key) -> str:
        target = self._gen_select_target(term.original)
        selector = "data#>" if not term.func else "data#>>"
        selection = f"{selector}'{{{target}}}'"
        return self._maybe_apply_function(term, selection)

    def _gen_sql_array_selection(self, term: SelectTerm, parsed: ArraySpecific) -> str:
        target = self._gen_select_target(term.bare_term)
        indexer = "->" if not term.func else "->>"
        selection = f"""
            case when data#>'{{{target}}}'{indexer}{parsed.idx} is not null then
                data#>'{{{target}}}'{indexer}{parsed.idx}
            else null end
            """
        return self._maybe_apply_function(term, selection)

    def _gen_sql_array_sub_selection(
        self,
        term: SelectTerm,
        parsed: Union[
            ArraySpecificSingle,
            ArraySpecificMultiple,
            ArrayBroadcastSingle,
            ArrayBroadcastMultiple,
        ],
    ) -> str:
        target = self._gen_select_target(term.bare_term)
        sub_selections = ','.join(parsed.sub_selections)
        data_selection_expr = f"filter_array_elements(data#>'{{{target}}}','{{{sub_selections}}}')"
        if (
            isinstance(parsed, ArraySpecificSingle)
            or isinstance(parsed, ArraySpecificMultiple)
        ):
            data_selection_expr = f'{data_selection_expr}->{parsed.idx}'
        selection = f"""
            case
                when data#>'{{{target}}}' is not null
                and jsonb_typeof(data#>'{{{target}}}') = 'array'
            then {data_selection_expr}
            else null end
            """
        return self._maybe_apply_function(term, selection)

    def _gen_sql_col(self, term: Union[SelectTerm, WhereTerm, OrderTerm]) -> str:
        if isinstance(term, WhereTerm) or isinstance(term, OrderTerm):
            select_term = term.parsed[0].select_term
        elif isinstance(term, SelectTerm):
            select_term = term
        if isinstance(term, WhereTerm):
            final_select_op = '#>>' # due to integer comparisons
        else:
            final_select_op = '#>'
        if len(select_term.parsed) > 1:
            test_select_term = select_term.parsed[-1]
            if isinstance(test_select_term, ArraySpecific):
                target = self._gen_select_target(select_term.bare_term)
                _idx = select_term.parsed[-1].idx
                col = f"data#>'{{{target}}}'{final_select_op}'{{{_idx}}}'"
            elif isinstance(test_select_term, ArraySpecificSingle):
                target = self._gen_select_target(select_term.bare_term)
                _idx = select_term.parsed[-1].idx
                _col = select_term.parsed[-1].sub_selections[0]
                col = f"data#>'{{{target}}}'#>'{{{_idx}}}'#>'{{{_col}}}'"
            else:
                target = self._gen_select_target(select_term.bare_term)
                col = f"data{final_select_op}'{{{target}}}'"
        else:
            if not (
                isinstance(select_term.parsed[0], Key) or
                isinstance(select_term.parsed[0], ArraySpecific)
            ):
                raise ParseError(f'Unsupported where term: {term.original}')
            if isinstance(select_term.parsed[0], ArraySpecific):
                target = self._gen_select_target(select_term.bare_term)
                _idx = select_term.parsed[0].idx
                col = f"data#>'{{{target}}}'{final_select_op}'{{{_idx}}}'"
            else:
                target = select_term.parsed[0].element
                col = f"data{final_select_op}'{{{target}}}'"
        if isinstance(term, WhereTerm):
            try:
                integer_ops = ['gt', 'gte', 'lt', 'lte']
                int(term.parsed[0].val)
                if (
                    term.parsed[0].op in integer_ops
                    and str(float(term.parsed[0].val)) != str(term.parsed[0].val)
                ):
                    col = f'({col})::int'
                elif str(float(term.parsed[0].val)) == str(term.parsed[0].val):
                    col = f'({col})::real'
            except ValueError:
                pass
        return col

    def _gen_sql_update(self, term: SetTerm) -> str:
        key = term.parsed[0].select_term.bare_term
        if not self.data or key not in self.data.keys():
            raise ParseError(f'Target key of update: {key} not found in payload')
        val = json.dumps(self.data[key]).replace("'", "''") # to handle single quotes inside
        return f" set data = jsonb_set(data, '{{{key}}}', ('{val}')::jsonb)"

    def _gen_select_with_retention(self, backup_cutoff: str) -> str:
        return f"(select * from {self.table_name} where data->>'timestamp' >= '{backup_cutoff}')a"

    def _gen_array_agg(self, query: str) -> str:
        return f"select json_agg(data) from ({query})a"

    def _maybe_float(self, val: Any) -> Union[str, float]:
        return f"'{val}'"




from typing import Optional

from pysquril.backends import SqliteBackend, sqlite_init, PostgresBackend

class B(object):

    """
    A backend for running interactive pysquril select queries -
    a tool for understanding the query language and the library
    capabilities.

    To explore queries with ephemeral sqlite backend:

        B().D(
            [
                {"x": 0, "y": 1},
                {"x": 100, "y": 4, "z": [1,2]}
            ]
        ).Q("select=x,z[1]&order=x.desc")

    -> [[100, 2], [0, None]]

    Broadcasting queries across tables with asterisk matching on table
    names, with a persistent sqlite DB:

        B(sqlite_path="/tmp/lol.db").T("t1").D([{"a":4}])
        B(sqlite_path="/tmp/lol.db").T("t2").D([{"a":4}])

        B(sqlite_path="/tmp/lol.db").T("t*").Q("select=a")

        -> [{'t1': ['[4]']}, {'t2': ['[4]']}]

    Or with a postgres backend, using lists of tables:

        eninge = postgres_init(...)
        be = PostgresBackend(engine)
        B(backend=be).T("t1").D([{"x": 1}])
        B(backend=be).T("t2").D([{"x": 1}])

        B(backend=be).T("t1,t2").Q("select=x")

        -> [{'t1': [[1]]}, {'t2': [[1]]}]


    """

    def __init__(
        self,
        backend: Optional[PostgresBackend] = None,
        sqlite_path: Optional[str] = ":memory:",
        verbose: bool = False,
    ):
        if not backend:
            engine = sqlite_init(sqlite_path)
            self.backend = SqliteBackend(engine)
        else:
            self.backend = backend
        self.verbose = verbose
        self._table_name = "temp"

    @property
    def table_name(self) -> str:
        return self._table_name

    def T(self, table_name: str):
        """
        Define the table name.

        """
        self._table_name = table_name
        return self

    def D(self, data: list):
        """
        Pass data to the backend.

        """
        self.backend.table_insert(
            table_name=self.table_name, data=data,
        )
        return self

    def Q(self, query: str) -> tuple:
        """
        Run a select query, print the results,
        return the query and the results.

        """
        if self.verbose:
            print(query)
        result = list(
            self.backend.table_select(
                table_name=self.table_name, uri_query=query,
            )
        )
        print(result)
        return query, result




AUDIT_SEPARATOR = "_"
AUDIT_SUFFIX = "audit"

def audit_table(table_name: str) -> str:
    """
    Construct audit table name from table_name.

    """
    return table_name + AUDIT_SEPARATOR + AUDIT_SUFFIX

def audit_table_src(audit_table_name: str) -> str:
    """
    Construct name of source table for a given audit table.

    """
    return audit_table_name[:-(len(AUDIT_SEPARATOR + AUDIT_SUFFIX))]




import datetime
import json
import logging
import sqlite3
import uuid

from abc import ABC, abstractmethod
from contextlib import contextmanager
from datetime import timedelta
from typing import Union, ContextManager, Iterable, Optional, Any, Callable
from urllib.parse import unquote
from uuid import uuid4

import psycopg2
import psycopg2.extensions
import psycopg2.pool

from pysquril.exc import DataIntegrityError, ParseError, OperationNotPermittedError
from pysquril.generator import SqliteQueryGenerator, PostgresQueryGenerator
from pysquril.utils import audit_table, audit_table_src, AUDIT_SEPARATOR, AUDIT_SUFFIX

def sqlite_init(path: str) -> sqlite3.Connection:
    engine = sqlite3.connect(path)
    return engine


def postgres_init(dbconfig: dict) -> psycopg2.pool.SimpleConnectionPool:
    min_conn = 2
    max_conn = 5
    dsn = f"dbname={dbconfig['dbname']} user={dbconfig['user']} password={dbconfig['pw']} host={dbconfig['host']}"
    pool = psycopg2.pool.SimpleConnectionPool(
        min_conn, max_conn, dsn
    )
    return pool


@contextmanager
def sqlite_session(
    engine: sqlite3.Connection,
) -> ContextManager[sqlite3.Cursor]:
    session = engine.cursor()
    try:
        yield session
        session.close()
    except Exception as e:
        session.close()
        engine.rollback()
        raise e
    finally:
        session.close()
        engine.commit()


@contextmanager
def postgres_session(
    pool: psycopg2.pool.SimpleConnectionPool,
) -> ContextManager[psycopg2.extensions.cursor]:
    engine = pool.getconn()
    session = engine.cursor()
    try:
        yield session
        session.close()
    except Exception as e:
        session.close()
        engine.rollback()
        raise e
    finally:
        session.close()
        engine.commit()
        pool.putconn(engine)


class AuditTransaction(object):

    """
    Container for generating audit events.
    Keeps state for transaction IDs, generates
    timestamps, and event IDs, propagates
    audit messages.

    There are five types of audit events:

    1. update - changes to existing data (default on)
    2. delete - deletions of existing data (default on)
    3. restore - rolling back update and/or delete events (default on)
    4. create - records of new data (default off)
    5. read - who accessed with a given query  (default off)

    Notes:

    If create and read events are used, one could calculate
    who looked at data for a given data subject, when on the
    basis of audit data alone. One would have to apply the
    query in each read event to the state of the data at the
    time of the query, and check if the data belonging to the
    subject (and/or the identfier of the data subject)
    is contained in the returned result.

    """

    def __init__(
        self,
        identity: str,
        message: Optional[str] = "",
        identity_name: Optional[str] = None,
    ) -> None:
        self.identity = identity
        self.identity_name = identity_name
        self.timestamp = datetime.datetime.now().isoformat()
        self.transaction_id = self._id()
        self.message = message

    def _id(self) -> str:
        return str(uuid4())

    def _event(self, diff: Any, previous: Any, event: str, query: str) -> dict:
        return {
            "diff": diff,
            "previous": previous,
            "event": event,
            "timestamp": self.timestamp,
            "identity": self.identity,
            "identity_name": self.identity_name,
            "event_id": self._id(),
            "transaction_id": self.transaction_id,
            "query": query,
            "message": self.message,
        }

    def event_update(self, *, diff: Any, previous: Any, query: str) -> dict:
        return self._event(diff, previous, "update", query)

    def event_delete(self, *, diff: Any, previous: Any, query: str) -> dict:
        return self._event(diff, previous, "delete", query)

    def event_restore(self, *, diff: Any, previous: Any, query: str) -> dict:
        return self._event(diff, previous, "restore", query)

    def event_create(self, *, diff: Any) -> dict:
        return self._event(diff, None, "create", None)

    def event_read(self, *, query: str) -> dict:
        return self._event(None, None, "read", query)


class DatabaseBackend(ABC):

    sep: str # schema separator character
    generator_class: Union[SqliteQueryGenerator, PostgresQueryGenerator]
    json_object_func: str

    def __init__(
        self,
        engine: Union[
            sqlite3.Connection,
            psycopg2.pool.SimpleConnectionPool,
        ],
        schema: str = None,
        verbose: bool = False,
        requestor: str = None,
        backup_days: Optional[int] = None,
        schema_pattern: Optional[str] = None,
        requestor_name: Optional[str] = None,
    ) -> None:
        super(DatabaseBackend, self).__init__()
        self.engine = engine
        self.verbose = verbose
        self.requestor = requestor
        self.requestor_name = requestor_name
        self.backup_days = backup_days
        self.schema_pattern = schema_pattern

    @abstractmethod
    def initialise(self) -> Optional[bool]:
        pass

    @abstractmethod
    def tables_list(self) -> list:
        pass

    @abstractmethod
    def table_create(
        self,
        table_name: str,
        session: Union[sqlite3.Cursor, psycopg2.extensions.cursor],
    ) -> bool:
        pass

    @abstractmethod
    def table_insert(
        self,
        table_name: str,
        data: Union[dict, list],
        session: Optional[Union[sqlite3.Cursor, psycopg2.extensions.cursor]] = None,
        audit: bool = False,
    ) -> bool:
        pass

    @abstractmethod
    def table_update(
        self,
        table_name: str,
        uri_query: str,
        data: dict,
        tsc: Optional[AuditTransaction] = None,
        session: Optional[Union[sqlite3.Cursor, psycopg2.extensions.cursor]] = None,
    ) -> bool:
        pass

    @abstractmethod
    def table_delete(self, table_name: str, uri_query: str, audit: bool = True) -> bool:
        pass

    @abstractmethod
    def table_select(
        self,
        table_name: str,
        uri_query: str,
        data: Optional[Union[dict, list]] = None,
        audit: bool = False,
    ) -> Iterable[tuple]:
        pass

    @abstractmethod
    def table_restore(self, table_name: str, uri_query: str) -> bool:
        pass

    @abstractmethod
    def table_alter(self, table_name: str, uri_query: str) -> dict:
        pass


class GenericBackend(DatabaseBackend):

    """Implementation of common methods for specific backends."""

    def _session_func(self) -> Callable:
        raise NotImplementedError

    def _diff_entries(self, current_entry: dict, target_entry: dict) -> dict:
        """
        Calculate the difference between two dictionaries, show the difference
        between the previous relative to current entry. E.g.:

        _diff_entries({a: 3, b: 4}, {a: 3, b: 5}) -> {b: 5}

        The diff is recorded in the new audit log when moving the state
        from current to previous.

        """
        out = {}
        for k, v in target_entry.items():
            if current_entry.get(k) != v:
                out[k] = v
        return out

    def _get_pk_value(self, primary_key: str, entry: dict) -> Any:
        keys = primary_key.split(".")
        if len(keys) == 1:
            return entry.get(primary_key)
        else:
            for key in keys:
                nested_result = entry.get(key)
                entry = nested_result
            return nested_result

    def _audit_source_exists(self, table_name: str) -> bool:
        """
        Check if the table from which audit records originate
        still exists.

        """
        exists = False
        try:
            current_data = list(self.table_select(audit_table_src(table_name), ""))
            exists = True
        except (sqlite3.OperationalError, psycopg2.errors.UndefinedTable):
            pass
        return exists

    def table_restore(self, table_name: str, uri_query: str) -> dict:
        """
        Restore rows to previous states, as recorded in the audit log.

        This can be either: undoing a delete, or updating a current
        record to a state prior to a specific update. The desired
        state is specified by means of a URI query. If the query
        yields multiple states of the same row, then the restore
        function will choose the oldest state, and ignore newer ones.

        Clients must provide the primary key which is used
        as the unique identifier, so we are able to identify the current
        row to which we need to apply changes.

        Examples:

        - restore one or more rows to a state prior to a specific
          update call:

            ?restore&where=transaction_id=eq.uuid&primary_key=key_name

        - restore a specific row to a prior state:

            ?restore&where=event_id=eq.uuid&primary_key=key_name

        - restore all rows to their first state after insert, before
          update or deletion:

            ?restore&primary_key=key_name

        When deleted rows are restored the event is recorded in the
        audit log as "restore" events. Such events are ignored by
        this function: that is, restore events are not restored,
        only updates and deletes.

        """
        work_done = {"restores": [], "updates": []}
        # ensure we have enough information
        query_parts = uri_query.split("&")
        if not query_parts:
            raise ParseError("Missing query")
        if "restore" not in query_parts:
            raise ParseError("Missing restore directive")
        has_pk = False
        message = ""
        for part in query_parts:
            if part.startswith("primary_key="):
                primary_key = part.split("=")[-1]
                if primary_key != "":
                    has_pk = True
            if part.startswith("message="):
                message = unquote(part.split("=")[-1])
        if not has_pk:
            raise ParseError("Missing primary_key")
        # fetch a copy of the current state, and all primay keys
        table_exists = False
        try:
            current_data = list(self.table_select(table_name, ""))
            current_pks = list(self.table_select(table_name, f"select={primary_key}"))
            table_exists = True
        except (sqlite3.OperationalError, psycopg2.errors.UndefinedTable):
            current_data = []
            current_pks = []
        # fetch the desired state
        if "order" in query_parts:
            uri_query = uri_query.split("&order")[0]
        uri_query = f"{uri_query}&order=timestamp.asc" # sorted from old to new
        target_data = list(self.table_select(audit_table(table_name), uri_query))
        if not target_data:
            return work_done # nothing to do
        tsc = AuditTransaction(self.requestor, message, self.requestor_name)
        session_func = self._session_func()
        try:
            # ensure the table exists, may have been deleted
            with session_func(self.engine) as session:
                self.table_create(table_name, session)
        except psycopg2.errors.DuplicateObject as e:
            pass # already exists
        handled = []
        with session_func(self.engine) as session:
            for entry in target_data:
                target_entry = entry.get("previous")
                pk_value = self._get_pk_value(primary_key, target_entry) if target_entry else None
                if pk_value in handled or entry.get("event") in ["restore", "create", "read"]:
                    continue
                target_entry = entry.get("previous")
                result = list(
                    self.table_select(
                        table_name,
                        f"where={primary_key}=eq.{pk_value}",
                    )
                )
                if len(result) > 1:
                    raise DataIntegrityError(f"primary_key: {primary_key} is not unique")
                elif not result:
                    # then it is currently deleted
                    self.table_insert(table_name, target_entry, session)
                    self.table_insert(
                        audit_table(table_name),
                        tsc.event_restore(diff=target_entry, previous=None, query=uri_query),
                        session,
                    )
                    work_done["restores"].append(entry)
                else:
                    current_entry = result[0]
                    diff = self._diff_entries(current_entry, target_entry)
                    if diff:
                        # note: query construction depends on the constraint
                        # that only top-level keys are allowed in set operations
                        self.table_update(
                            table_name,
                            f"set={','.join(diff.keys())}&where={primary_key}=eq.{pk_value}",
                            data=diff,
                            tsc=tsc,
                            session=session,
                        )
                        work_done["updates"].append(entry)
                handled.append(pk_value)
        return work_done

    def _tables_in_schemas(self, table_name: str) -> list:
        """
        Return a list of all existing instances of the {table_name}
        across all schemas, including the schema name: {schema}.{table_name}.

        """
        raise NotImplementedError

    def _create_all_view(
        self,
        view_name: str,
        unions: str,
        session: Union[sqlite3.Cursor, psycopg2.extensions.cursor],
    ) -> None:
        """
        Backend specific implementation of view creation.

        """
        raise NotImplementedError

    def _fqtn(self, table_name: str, schema_name: Optional[str] = None, no_schema: bool = False) -> str:
        """
        Return a fully qualified table name - qualified with the schema.

        """
        raise NotImplementedError

    def _define_all_view(self, table_name: str) -> None:
        """
        To allow queries across all schemas, for a given table.

        The method is optionally called as part of table_insert,
        and table_delete - when a new table is created, and when
        an existing table is deleted.

        It works like this: suppose there are no tables at all.
        The first insert is done into table p11.A, and
        _define_all_view(A) is called. This creates a view named
        all.A - a view defined in the schema named 'all', with
        the name of the table which is being created. The view
        returns all data from tables named A in all schemas. At present
        it would be: create or replace view all.A as select * from p11.A;.

        Now the first insert is done into a new table in another schema p12 ->
        _define_all_view(A) is called. Now the view called
        all.A is updated to be the union of p11.A and p12.A:
        create or replace view all.A as select * from p11.A
        union all select * from p12.A;.

        Upon deletion of table p12.A, the view definition is
        updated to remove the reference to the deleted table.

        With the view in place, a caller can do table_select(all.{table_name})
        to query a given table across all schemas.

        """
        tables = self._tables_in_schemas(table_name)
        if not tables:
            return # when the last one is deleted
        unions = " union all ".join(
            [f"select * from {t}" for t in tables]
        )
        view_name = self._fqtn(table_name, schema_name="all")
        session_func = self._session_func()
        with session_func(self.engine) as session:
            self._create_all_view(view_name, unions, session)

    def _query_for_select(
        self,
        table_name: str,
        uri_query: str,
        data: Optional[Union[dict, list]] = None,
        array_agg: bool = False,
        apply_cutoff: bool = False,
    ) -> str:
        """
        Return the appropriate select statement for a given
        table_name, and uri_query, calculating any backup
        cutoff for audit data if needed.

        """
        backup_cutoff = None
        if apply_cutoff:
            backup_cutoff = (datetime.date.today() - timedelta(days=self.backup_days)).isoformat()
        sql = self.generator_class(
            f'{self._fqtn(table_name)}',
            uri_query,
            data=data,
            backup_cutoff=backup_cutoff,
            array_agg=array_agg,
        )
        return sql.select_query

    def _query_for_select_many(self, uri_query: str, tables: list, apply_cutoff: bool = False) -> str:
        queries = []
        for table_name in tables:
            sql = self._query_for_select(table_name, uri_query, array_agg=True, apply_cutoff=apply_cutoff)
            queries.append(f"select {self.json_object_func}('{table_name}', ({sql}))")
        return " union all ".join(queries)

    def _yield_results(self, query: str)-> Iterable[tuple]:
        raise NotImplementedError

    def _is_audit_table(self, table_name: str) -> bool:
        """
        Determine whether a given table is an audit table.

        """
        sufficient = False
        neccesary = table_name.endswith(AUDIT_SEPARATOR + AUDIT_SUFFIX)
        if not neccesary:
            return neccesary and sufficient
        try:
            dummy_event = AuditTransaction("").event_read(query="")
            result = list(
                self._yield_results(
                    f"select data from {table_name} limit 1"
                )
            )[0]
            sufficient = (
                uuid.UUID(result.get("transaction_id"))
                and uuid.UUID(result.get("event_id"))
                and result.get("event") in ["update", "delete", "read", "create"]
                and result.get("timestamp") is not None
                and set(result.keys()).difference(dummy_event.keys()) == set()
            )

        except Exception as e:
            pass
        return neccesary and sufficient

    def table_select(
        self,
        table_name: str,
        uri_query: str,
        data: Optional[Union[dict, list]] = None,
        exclude_endswith: list = [],
        audit: bool = False,
    ) -> Iterable[tuple]:
        """
        Yield a resulset associated with a table_name, and a uri_query.

        The table_name can be either a reference to a specific table, or an
        asterisk expression intended to match a set of table names. In the
        latter case, a query is constructed to union the resulsets from
        all the relevant tables together.

        Optionally exclude tables that end with a specific pattern.

        """
        apply_cutoff = (
            self._is_audit_table(table_name)
            and not self._audit_source_exists(table_name)
            and self.backup_days is not None
        )
        if "*" in table_name:
            tables = self.tables_list(exclude_endswith = exclude_endswith, table_like=table_name)
            if not tables:
                return iter([])
            query = self._query_for_select_many(uri_query, tables, apply_cutoff=apply_cutoff)
        elif "," in  table_name:
            tables = table_name.split(",")
            query = self._query_for_select_many(uri_query, tables, apply_cutoff=apply_cutoff)
        else:
            query = self._query_for_select(table_name, uri_query, data, apply_cutoff=apply_cutoff)
        if audit:
            tsc = AuditTransaction(identity=self.requestor, identity_name=self.requestor_name)
            self.table_insert(audit_table(table_name), tsc.event_read(query=uri_query))
        return self._yield_results(query)

    def table_delete(
        self,
        table_name: str,
        uri_query: str,
        update_all_view: Optional[bool] = False,
        audit: bool = True,
    ) -> bool:
        """
        Delete the intended data from the table if a uri_query
        is present, otherwise drop the table. All deletions
        are recorded in the audit log by default, but this can
        be disabled by callers. This is useful for cases where
        data is being deleted completely for compliance purposes.

        """
        audit_data = []
        sql = self.generator_class(f'{self._fqtn(table_name)}', uri_query)
        if audit:
            tsc = AuditTransaction(self.requestor, sql.message, self.requestor_name)
            for row in self.table_select(table_name, uri_query):
                audit_data.append(tsc.event_delete(diff=None, previous=row, query=uri_query))
        with self._session_func()(self.engine) as session:
            session.execute(sql.delete_query)
            if not table_name.endswith("_audit") and audit:
                self.table_create(audit_table(table_name), session)
                self.table_insert(audit_table(table_name), audit_data, session)
        if update_all_view:
            self._define_all_view(table_name)
        return True

    def table_update(
        self,
        table_name: str,
        uri_query: str,
        data: dict,
        tsc: Optional[AuditTransaction] = None,
        session: Optional[Union[sqlite3.Cursor, psycopg2.extensions.cursor]] = None,
    ) -> bool:
        """
        Update one or more keys, recording changes in the audit log.

        """
        if self._is_audit_table(table_name):
            raise OperationNotPermittedError("audit tables cannot be altered directly")
        audit_data = []
        sql = self.generator_class(f'{self._fqtn(table_name)}', uri_query, data=data)
        tsc = AuditTransaction(self.requestor, sql.message, self.requestor_name) if not tsc else tsc
        for val in self.table_select(table_name, uri_query, data=data):
            audit_data.append(tsc.event_update(diff=data, previous=val, query=uri_query))
        if session:
            session.execute(sql.update_query)
            self.table_insert(audit_table(table_name), audit_data, session)
        else:
            with self._session_func()(self.engine) as session:
                session.execute(sql.update_query)
            self.table_insert(audit_table(table_name), audit_data)
        return True

    def table_alter(self, table_name: str, uri_query: str) -> dict:
        """
        Alter the name of a table, and its audit table (if it exists).
        Return information about which tables were altered.

        """
        if self._is_audit_table(table_name):
            raise OperationNotPermittedError("audit tables cannot be altered directly")
        sql = self.generator_class(
            f'{self._fqtn(table_name)}',
            uri_query,
            table_name_func=self._fqtn,
        )
        with self._session_func()(self.engine) as session:
            session.execute(sql.alter_query)
        altered = {"tables": [table_name]}
        try:
            audit_table_name = audit_table(table_name)
            sql = self.generator_class(
                f'{self._fqtn(audit_table_name)}',
                uri_query,
                table_name_func=self._fqtn,
                audit=True,
            )
            with self._session_func()(self.engine) as session:
                session.execute(sql.alter_query)
            altered["tables"].append(audit_table_name)
        except (psycopg2.errors.UndefinedTable, sqlite3.OperationalError):
            pass
        return altered

    def _audit_insert(self, table_name: str, data: Union[str, list]) -> bool:
        tsc = AuditTransaction(identity=self.requestor, identity_name=self.requestor_name)
        audit_data = []
        if type(data) is list:
            for row in data:
                audit_data.append(tsc.event_create(diff=row))
        else:
            audit_data = [tsc.event_create(diff=data)]
        self.table_insert(audit_table(table_name), audit_data)
        return True


class SqliteBackend(GenericBackend):

    """
    This backend works reliably, and offers decent read-write
    performance to API clients under the following conditions:

    a) using network storage, like NFS

        - using rollback-mode for transactions
        - where clients do not perform long-running read operations

    b) not using network storage

        - using WAL-mode for transactions
        - clients can perform long-running reads without
          blocking writers

    Briefly, WAL-mode (which offer non-blocking read/write) cannot
    be reliably used over NFS, because the locking primitives
    used by SQLite to prevent DB corruption are not implemented. So
    if you are using NFS, you must use rollback-mode. This however,
    means that writers require exclusive locks to write, which means
    that long-running reads by clients would block writers. This
    may be unfortunate, depending on the use case.

    For more background refer to:

        - https://www.sqlite.org/lockingv3.html
        - https://www.sqlite.org/wal.html
        - https://www.sqlite.org/threadsafe.html

    """

    generator_class = SqliteQueryGenerator
    json_object_func = 'json_object'

    def __init__(
        self,
        engine: sqlite3.Connection,
        verbose: bool = False,
        schema: str = None,
        requestor: str = None,
        backup_days: Optional[int] = None,
        schema_pattern: Optional[str] = None,
        requestor_name: Optional[str] = None,
    ) -> None:
        self.engine = engine
        self.verbose = verbose
        self.table_definition = '(data json unique not null)'
        self.schema = schema if schema else ""
        self.sep = "_" if self.schema else ""
        self.requestor = requestor
        self.requestor_name = requestor_name
        self.backup_days = backup_days
        self.schema_pattern = schema_pattern

    def _session_func(self) -> Callable:
        return sqlite_session

    def _fqtn(self, table_name: str, schema_name: Optional[str] = None, no_schema: bool = False) -> str:
        """
        Return a fully qualified table name - qualified with the schema.

        """
        schema = schema_name or self.schema
        return f'"{schema}{self.sep}{table_name}"'

    def _tables_in_schemas(self, table_name: str) -> list:
        """
        Return a list of all existing instances of the {table_name}
        across all schemas, including the schema name: {schema}.{table_name}.

        """
        with sqlite_session(self.engine) as session:
            res = session.execute(
                f"""select name FROM sqlite_master where type = 'table'
                    and name like '{self.schema_pattern}%{table_name}'
                """
            ).fetchall()
        return [ r[0] for r in res ] if res else []

    def _create_all_view(self, view_name: str, unions: str, session: sqlite3.Cursor) -> None:
        session.execute(f'drop view if exists {view_name}')
        session.execute(f'create view {view_name} as {unions}')

    def initialise(self) -> Optional[bool]:
        pass

    def tables_list(
        self,
        exclude_endswith: list = [],
        only_endswith: Optional[str] = None,
        remove_pattern: Optional[str] = None,
        table_like: Optional[str] = "",
    ) -> list:
        table_like_filter = ""
        if table_like:
            pattern = table_like.replace("*", "%")
            table_like_filter = f"and name like '{pattern}'"
        query = f"select name FROM sqlite_master where type = 'table' {table_like_filter} order by name asc"
        with sqlite_session(self.engine) as session:
            res = session.execute(query).fetchall()
        if not res:
            return []
        else:
            out = []
            for row in res:
                name = row[0]
                exclude = False
                if only_endswith:
                    if not name.endswith(only_endswith):
                        exclude = True
                for ends_with in exclude_endswith:
                    if name.endswith(ends_with):
                        exclude = True
                if not exclude:
                    name = name.replace(remove_pattern, "") if remove_pattern else name
                    out.append(name)
            return out

    def table_create(
        self,
        table_name: str,
        session: sqlite3.Cursor,
    ) -> bool:
        session.execute(f'create table if not exists {self._fqtn(table_name)} {self.table_definition}')
        return True

    def table_insert(
        self,
        table_name: str,
        data: Union[dict, list],
        session: Optional[sqlite3.Cursor] = None,
        update_all_view: Optional[bool] = False,
        audit: bool = False,
    ) -> bool:
        try:
            dtype = type(data)
            insert_stmt = f'insert into {self._fqtn(table_name)} (data) values (?)'
            target = []
            if dtype is list:
                for element in data:
                    target.append((json.dumps(element),))
            elif dtype is dict:
                target.append((json.dumps(data),))
            if session:
                # in this case we are re-using a session
                # from a context manager estabilshed by the caller
                # and if an exception is raised, the caller handles it
                session.executemany(insert_stmt, target)
            else:
                try:
                    with sqlite_session(self.engine) as session:
                        session.executemany(insert_stmt, target)
                except (sqlite3.ProgrammingError, sqlite3.OperationalError) as e:
                    with sqlite_session(self.engine) as session:
                        self.table_create(table_name, session)
                        session.executemany(insert_stmt, target)
                    if update_all_view:
                        self._define_all_view(table_name)
            if audit:
                self._audit_insert(table_name, data)
            return True
        except sqlite3.IntegrityError as e:
            logging.info('Ignoring duplicate row')
            return True # idempotent PUT
        except sqlite3.ProgrammingError as e:
            logging.error('Syntax error?')
            raise e
        except sqlite3.OperationalError as e:
            logging.error('Database issue')
            raise e
        except Exception as e:
            logging.error('Not sure what went wrong')
            raise e

    def _yield_results(self, query: str)-> Iterable[tuple]:
        with sqlite_session(self.engine) as session:
            for row in session.execute(query):
                yield json.loads(row[0])


class PostgresBackend(GenericBackend):

    """
    A PostgreSQL backend. PostgreSQL is a full-fledged
    client-server relational DB, implementing MVCC for
    concurrency control. This backend is therfore
    suitable for applications that have many simultaneous
    read and write operations, with full ACID compliance
    and no reader-writer blocking.

    For more info on MVCC, see:
    https://www.postgresql.org/docs/12/mvcc-intro.html

    """

    generator_class = PostgresQueryGenerator
    sep = "."
    json_object_func = 'jsonb_build_object'

    def __init__(
        self,
        pool: psycopg2.pool.SimpleConnectionPool,
        verbose: bool = False,
        schema: str = None,
        requestor: str = None,
        backup_days: Optional[int] = None,
        schema_pattern: Optional[str] = None,
        requestor_name: Optional[str] = None,
    ) -> None:
        self.engine = pool
        self.verbose = verbose
        self.table_definition = '(data jsonb not null, uniq text unique not null)'
        self.schema = schema if schema else 'public'
        self.requestor = requestor
        self.requestor_name = requestor_name
        self.backup_days = backup_days
        self.schema_pattern = schema_pattern

    def _session_func(self) -> Callable:
        return postgres_session

    def _fqtn(self, table_name: str, schema_name: Optional[str] = None, no_schema: bool = False) -> str:
        """
        Return a fully qualified table name - qualified with the schema.

        """
        if no_schema:
            return f'"{table_name}"'
        schema = schema_name or self.schema
        schema = '"all"' if schema == "all" else schema # all is a reserved word
        return f'{schema}{self.sep}"{table_name}"'

    def _tables_in_schemas(self, table_name: str) -> list:
        """
        Return a list of all existing instances of the {table_name}
        across all schemas, including the schema name: {schema}.{table_name}.

        """
        with postgres_session(self.engine) as session:
            session.execute(
                f"""select concat_ws('.', table_schema, concat('"', table_name, '"'))
                    from information_schema.tables where table_schema
                    like '{self.schema_pattern}%' and table_name = '{table_name}'
                """
            )
            res = session.fetchall()
        return [ r[0] for r in res ] if res else []

    def _create_all_view(self, view_name: str, unions: str, session: psycopg2.extensions.cursor) -> None:
        session.execute(f'create schema if not exists "all"')
        session.execute(f"create or replace view {view_name} as {unions}")

    def initialise(self) -> Optional[bool]:
        try:
            with postgres_session(self.engine) as session:
                for stmt in self.generator_class.db_init_sql:
                        session.execute(stmt)
        except psycopg2.InternalError as e:
            pass # throws a tuple concurrently updated when restarting many processes
        return True

    def tables_list(
        self,
        exclude_endswith: list = [],
        only_endswith: Optional[str] = None,
        remove_pattern: Optional[str] = None,
        table_like: Optional[str] = "",
    ) -> list:
        table_like_filter = ""
        if table_like:
            pattern = table_like.replace("*", "%")
            table_like_filter = f"and table_name like '{pattern}'"
        query = f"""select table_name from information_schema.tables
            where table_schema = '{self.schema}' {table_like_filter} order by table_name asc"""
        with postgres_session(self.engine) as session:
            session.execute(query)
            res = session.fetchall()
        if not res:
            return []
        else:
            out = []
            for row in res:
                name = row[0]
                exclude = False
                if only_endswith:
                    if not name.endswith(only_endswith):
                        exclude = True
                for ends_with in exclude_endswith:
                    if name.endswith(ends_with):
                        exclude = True
                if not exclude:
                    name = name.replace(remove_pattern, "") if remove_pattern else name
                    out.append(name)
            return out

    def table_create(
        self,
        table_name: str,
        session: psycopg2.extensions.cursor,
    ) -> bool:
        table_create = f'create table if not exists {self._fqtn(table_name)}{self.table_definition}'
        trigger_create = f"""
            create trigger ensure_unique_data before insert
            on {self.schema}{self.sep}"{table_name}"
            for each row execute procedure unique_data()
        """ # change to create if not exists when pg ^v11
        session.execute( # need to check if table exists
            f"select exists(select from pg_tables where schemaname = '{self.schema}' and tablename = '{table_name}')"
        )
        exists = session.fetchall()[0][0]
        if not exists:
            session.execute(f'create schema if not exists {self.schema}')
            session.execute(table_create)
            session.execute(trigger_create)


    def table_insert(
        self,
        table_name: str,
        data: Union[dict, list],
        session: Optional[psycopg2.extensions.cursor] = None,
        update_all_view: Optional[bool] = False,
        audit: bool = False,
    ) -> bool:
        try:
            dtype = type(data)
            insert_stmt = f'insert into {self._fqtn(table_name)} (data) values (%s)'
            target = []
            if dtype is list:
                for element in data:
                    target.append((json.dumps(element),))
            elif dtype is dict:
                target.append((json.dumps(data),))
            if session:
                # in this case we are re-using a session
                # from a context manager estabilshed by the caller
                # and if an exception is raised, the caller handles it
                session.executemany(insert_stmt, target)
            else:
                try:
                    with postgres_session(self.engine) as session:
                        session.executemany(insert_stmt, target)
                except (psycopg2.ProgrammingError, psycopg2.OperationalError) as e:
                    with postgres_session(self.engine) as session:
                        self.table_create(table_name, session)
                        session.executemany(insert_stmt, target)
                    if update_all_view:
                        self._define_all_view(table_name)
            if audit:
                self._audit_insert(table_name, data)
            return True
        except psycopg2.IntegrityError as e:
            logging.info('Ignoring duplicate row')
            return True # idempotent PUT
        except psycopg2.ProgrammingError as e:
            logging.error('Syntax error?')
            raise e
        except psycopg2.OperationalError as e:
            logging.error('Database issue')
            raise e
        except Exception as e:
            logging.error('Not sure what went wrong')
            raise e

    def _yield_results(self, query: str)-> Iterable[tuple]:
        with postgres_session(self.engine) as session:
            session.execute(query)
            for row in session:
                yield row[0]




from http import HTTPStatus

class PySqurilError(Exception):
    status = HTTPStatus.BAD_REQUEST

    def __init__(self, reason: str = "") -> None:
        self.reason = reason

class ParseError(PySqurilError):
    status = HTTPStatus.BAD_REQUEST

class DataIntegrityError(PySqurilError):
    status = HTTPStatus.BAD_REQUEST

class OperationNotPermittedError(PySqurilError):
    status = HTTPStatus.BAD_REQUEST




import datetime
import json
import os
import sqlite3
import unittest
import uuid
import tempfile

from datetime import timedelta
from typing import Callable, Union
from urllib.parse import quote

import psycopg2
import psycopg2.errors
import psycopg2.extensions
import psycopg2.pool
import pytest

from termcolor import colored

from pysquril.backends import (
    SqliteBackend,
    PostgresBackend,
    sqlite_init,
    postgres_init,
    sqlite_session,
    postgres_session,
)
from pysquril.exc import ParseError, OperationNotPermittedError
from pysquril.generator import SqliteQueryGenerator, PostgresQueryGenerator
from pysquril.parser import (
    SelectClause,
    WhereClause,
    GroupByTerm,
    GroupByClause,
    AlterClause,
    UriQuery,
)
from pysquril.test_data import dataset
from pysquril.utils import audit_table


TEST_REQUESTOR = "p11-treq"
TEST_REQUESTOR_NAME = "Test Requestor"

class TestParser(object):

    def test_select(self) -> None:
        c = SelectClause("x[*|a,b],y.z")
        assert len(c.split_clause()) == 2

    def test_where(self) -> None:
        # TODO: test x=gt.3,or:(y=gt.3,and:z=not.is.null)

        # basic where's

        c = WhereClause("a=eq.b,and:lol=not.is.null")
        assert len(c.split_clause()) == 2
        c = WhereClause("a[1|h]=eq.0,or:b[0]=gt.1")
        assert len(c.split_clause()) == 2
        c = WhereClause("((x=gt.3,or:y=gt.3),and:z=not.is.null)")
        assert len(c.split_clause()) == 3

        # quoting values

        c = WhereClause("lol=eq.'.'")
        assert len(c.split_clause()) == 1
        where_term = c.parsed[0]
        where_element = where_term.parsed[0]
        assert where_element.select_term.bare_term == "lol"
        assert where_element.op == "eq"
        assert where_element.val == "."

        c = WhereClause("a=gte.4,and:b=eq.'r,[,],',or:c=neq.0")
        assert len(c.split_clause()) == 3

        # ampersand

        c = WhereClause("a=eq.'&'")
        assert len(c.split_clause()) == 1

        # escaping quotes

        c = WhereClause("lol=eq.'\\'n kat loop oor die pad'")
        where_term = c.parsed[0]
        where_element = where_term.parsed[0]
        assert where_element.val == "''n kat loop oor die pad"


    def test_group_by(self) -> None:

        with pytest.raises(ParseError):
            GroupByTerm("max(key)")

        with pytest.raises(ParseError):
            GroupByTerm("a[1|b,c]")

        with pytest.raises(ParseError):
            GroupByTerm("a[*|b]")

        with pytest.raises(ParseError):
            GroupByTerm("a[*|b,c]")

        c = GroupByClause("a,b")
        assert len(c.split_clause()) == 2

        c = GroupByClause("a.b.c,d")
        assert len(c.split_clause()) == 2

    def test_alter(self) -> None:

        c = AlterClause("name=eq.new_name")
        term = c.parsed[0]
        element = term.parsed[0]
        assert element.val == "new_name"

        with pytest.raises(ParseError):
            AlterClause("num=eq.new_name")

        with pytest.raises(ParseError):
            AlterClause("name=neq.new_name")

    def test_uri_query(self) -> None:

        q = UriQuery("", "")

        indices = q._index_clauses("x&'&'&y")
        assert indices == [1, 5]

        sliced = q._slice(target="1&a&bc", positions=[1, 3])
        assert sliced == ["1", "a", "bc"]

        q = UriQuery("table", "where=a=eq.'g\\'n mooi dag buite'")
        assert q.where.original == "a=eq.'g\\'n mooi dag buite'"
        assert q.where.parsed[0].parsed[0].val == "g''n mooi dag buite"


class TestBackends(object):

    verbose = True

    data = dataset

    def run_backend_tests(
        self,
        data: list,
        engine: Union[sqlite3.Connection, psycopg2.pool.SimpleConnectionPool],
        session_func: Callable,
        SqlGeneratorCls: Union[SqliteQueryGenerator, PostgresQueryGenerator],
        DbBackendCls: Union[SqliteBackend, PostgresBackend],
        verbose: bool,
    ) -> None:

        def run_select_query(
            uri_query: str,
            table:  str = 'test_table',
            engine: Union[sqlite3.Connection, psycopg2.pool.SimpleConnectionPool] = engine,
            verbose: bool = verbose,
        ) -> list:
            out = []
            if verbose:
                print(colored(uri_query, 'magenta'))
            q = SqlGeneratorCls(table, uri_query)
            if verbose:
                print(colored(q.select_query, 'yellow'))
            db = DbBackendCls(engine)
            out = list(db.table_select(table, uri_query))
            if verbose:
                print(out)
            return out

        def run_update_query(
            uri_query: str,
            table: str = 'test_table',
            engine: Union[sqlite3.Connection, psycopg2.pool.SimpleConnectionPool] = engine,
            verbose: bool = verbose,
            data: list = data,
        ) -> list:
            q = SqlGeneratorCls(table, uri_query, data=data)
            if verbose:
                print(colored(q.update_query, 'cyan'))
            db = DbBackendCls(engine)
            db.table_update(table, uri_query, data)
            out = list(db.table_select(table, ""))
            return out

        def run_delete_query(
            uri_query: str,
            table: str = 'test_table',
            engine: Union[sqlite3.Connection, psycopg2.pool.SimpleConnectionPool] = engine,
            verbose: bool = verbose,
        ) -> bool:
            q = SqlGeneratorCls(table, uri_query)
            if verbose:
                print(q.delete_query)
            db = DbBackendCls(engine)
            db.table_delete(table, uri_query)
            return True

        def run_alter_query(
            uri_query: str,
            table: str,
            engine: Union[sqlite3.Connection, psycopg2.pool.SimpleConnectionPool] = engine,
            verbose: bool = verbose,
        ) -> dict:
            db = DbBackendCls(engine)
            out = db.table_alter(table, uri_query)
            return out


        db = DbBackendCls(engine)
        try:
            db.table_delete('test_table', '')
        except Exception as e:
            pass
        try:
            db.table_delete('another_table', '')
        except Exception as e:
            pass
        try:
            db.table_delete('silly_table', '')
            db.table_delete(audit_table('silly_table'), '')
        except Exception as e:
            pass

        # test '*' without any tables
        out = list(db.table_select('*', 'select=count(1)', exclude_endswith = ['_audit', '_metadata']))
        assert list(out) == []

        # create tables
        db.table_insert('test_table', data)
        db.table_insert('another_table', data)

        # SELECT
        if verbose:
            print('\n===> SELECT\n')
        # simple key selection
        out = run_select_query('select=x')
        for entry in out:
            assert isinstance(entry, list)
        assert out[0][0] == 1900
        # more than one simple key
        out = run_select_query('select=x,z')
        assert len(out[0]) == 2
        # nested key
        out = run_select_query('select=a.k1')
        assert len(out[0]) == 1
        assert out[2] == [{'r1': [1, 2], 'r2': 2}]
        # simple array slice
        out = run_select_query('select=x,b[1]')
        assert out[0][1] == 2
        assert out[1][1] == 1
        # nested simple array slice
        out = run_select_query('select=x,a.k2[1]')
        assert out[2] == [88, 9]
        # selecting a key inside an array slice
        out = run_select_query('select=x,c[1|h]')
        assert out[0][1] is None
        assert out[1][1] == 32
        # selecting keys inside an array slice
        out = run_select_query('select=x,c[1|h,p]')
        assert out[0][1] is None
        assert out[1][1] == [32, 0] or out[1][1] == '[32,0]' # sqlite
        # broadcast key selection inside array - single key
        out = run_select_query('select=x,c[*|h]')
        assert len(out[1][1]) == 3
        assert out[1][1] == [3, 32, 0]
        # broadcast key selection inside array - mutliple keys
        out = run_select_query('select=x,c[*|h,p]')
        assert len(out[1][1]) == 3
        assert out[1][1][0] == [3, 99] or out[1][1][0] == '[3,99]' # sqlite
        # nested array selection
        out = run_select_query('select=a.k1.r1[0]')
        assert out[2] == [1]
        # nested keys
        # with single selection inside array, specific element
        out = run_select_query('select=a.k3[0|h]')
        assert out[3] == [0]
        # with single selection inside array, broadcast
        out = run_select_query('select=a.k3[*|h]')
        assert out[3] == [[0, 63]]
        # now multiple sub-selections
        out = run_select_query('select=a.k3[0|h,s]')
        assert out[3] == [[0, 521]] or out[3] == ['[0,521]'] # sqlite
        out = run_select_query('select=a.k3[*|h,s]')
        assert out[3] == [[[0, 521], [63, 333]]] or out[3] == [['[0,521]', '[63,333]']] # sqlite
        # multiple sub-keys
        out = run_select_query('select=a.k1,a.k3')
        assert out[3] == [{'r1': [33, 200], 'r2': 90}, [{'h': 0, 'r': 77, 's': 521}, {'h': 63, 's': 333}]]

        # FUNCTIONS/AGGREGATIONS
        # supported: count, avg, sum, (max, min), min_ts, max_ts
        if verbose:
            print('\n===> FUNCTIONS\n')

        out = run_select_query('select=count(1)')
        assert out == [[5]]
        out = run_select_query('select=count(*)')
        assert out == [[5]]
        out = run_select_query('select=count(x)')
        assert out == [[4]]
        out = run_select_query('select=count(1),min(y)')
        assert out == [[5, 1]]
        out = run_select_query('select=count(1),avg(x),min(y),sum(x),max_ts(timestamp)')
        assert out == [[5, 526.2500000000000000, 1, 2105, '2020-10-14T20:20:34.388511']]
        # nested selections
        out = run_select_query('select=count(a.k1.r2),count(x),count(*)')
        assert out == [[2, 4, 5]]
        # array selections
        out = run_select_query('select=count(b[0])')
        assert out == [[2]]
        out = run_select_query('select=max(b[0])')
        assert out == [[1111]]
        out = run_select_query('select=min_ts(timestamps[0])')
        assert out == [['1984-10-13T10:15:26.388573']]
        # sub-selections
        out = run_select_query('select=count(a.k3[0|h])')
        assert out == [[1]]
        out = run_select_query('select=max(q.r[0|s])')
        assert out == [[77]]

        if verbose:
            print('\n===> BROADCASTING\n')

        # broadcasting aggregations
        out = list(db.table_select('*', 'select=count(1)', exclude_endswith = ['_audit', '_metadata']))
        assert out == [{'another_table': [5]}, {'test_table': [5]}]

        # fuzzy matching
        out = list(db.table_select('another*', 'select=count(1)', exclude_endswith = ['_audit', '_metadata']))
        assert out == [{'another_table': [5]}]

        out = list(db.table_select('*_table', 'select=count(1)', exclude_endswith = ['_audit', '_metadata']))
        assert out == [{'another_table': [5]}, {'test_table': [5]}]

        # broadcasting queries without aggregation
        out = list(db.table_select('*', 'select=x', exclude_endswith = ['_audit', '_metadata']))
        assert out is not None
        assert len(out) == 2
        assert len(out[0].get("another_table")) == 5
        assert len(out[1].get("test_table")) == 5

        out = list(db.table_select('*', 'select=x,y&where=z=not.is.null', exclude_endswith = ['_audit', '_metadata']))
        assert out is not None
        assert len(out) == 2
        assert len(out[0].get("another_table")) == 4
        assert len(out[1].get("test_table")) == 4

        # table lists
        out = list(
            db.table_select(
                'another_table,test_table',
                'select=x,y&where=z=not.is.null',
                exclude_endswith = ['_audit', '_metadata']
            )
        )
        assert out is not None
        assert len(out) == 2
        assert len(out[0].get("another_table")) == 4
        assert len(out[1].get("test_table")) == 4

        # WHERE
        if verbose:
            print('\n===> WHERE\n')
        # simple key op
        out = run_select_query('where=x=gt.1000')
        assert out[0]['x'] == 1900
        # multipart simple key ops
        out = run_select_query('where=x=gt.1000,or:y=eq.11')
        assert len(out) == 2
        out = run_select_query('where=x=lt.1000,and:y=eq.11')
        assert out == []
        # groups (with a select)
        out = run_select_query('select=x&where=((x=lt.1000,and:y=eq.11),or:x=gt.1000)')
        assert out == [[1900]]
        # is, not, like, and null
        out = run_select_query('where=x=not.is.null')
        assert len(out) == 4
        out = run_select_query('select=d&where=d=not.like.*g3')
        assert len(out) == 2
        # Run the query with not as a string value.
        out=run_select_query('select=d&where=d=eq.not')
        assert len(out) == 0
        # in
        out = run_select_query('select=d&where=d=in.[string1,string2]')
        assert len(out) == 2
        out = run_select_query("select=contemplate&where=contemplate=in.['non arising','vanishing']")
        assert len(out) == 2
        assert ['non arising'] in out
        assert ['vanishing'] in out
        out = run_select_query("select=contemplate&where=contemplate=in.['g\\'n niks nie','vanishing']")
        assert len(out) == 2
        # nested key ops
        out = run_select_query('where=a.k1.r2=eq.90')
        assert len(out) == 1
        # nested key ops with slicing
        out = run_select_query('select=x&where=a.k1.r1[0]=eq.1')
        assert out[0][0] == 88
        out = run_select_query('select=x&where=a.k3[0|h]=eq.0')
        assert out[0][0] == 107
        # timestamps
        out = run_select_query('select=x,timestamp&where=timestamp=gt.2020-10-14')
        assert len(out) == 3
        out = run_select_query('select=x,timestamp&where=timestamp=lt.2020-10-14')
        assert len(out) == 2
        # equality with strings made of digits, and with integers
        out = run_select_query('select=x&where=lol1=eq.123')
        assert out[0][0] == 1900
        out = run_select_query('select=x&where=lol2=eq.123')
        assert out[0][0] == 1900
        out = run_select_query('select=x&where=lol3.yeah=eq.123')
        assert out[0][0] == 1900
        out = run_select_query('select=x&where=lol4.yeah=eq.123')
        assert out[0][0] == 1900
        # same as ^, but with non-equality, neq
        out = run_select_query('select=y&where=lol1=neq.123,and:lol1=not.is.null')
        assert out[0][0] == 11
        out = run_select_query('select=y&where=lol2=neq.123,and:lol1=not.is.null')
        assert out[0][0] == 11
        out = run_select_query('select=y&where=lol3.yeah=neq.123,and:lol1=not.is.null')
        assert out[0][0] == 11
        out = run_select_query('select=y&where=lol4.yeah=neq.123,and:lol1=not.is.null')
        assert out[0][0] == 11
        # floats
        out = run_select_query('select=z&where=float=eq.3.1')
        assert out[0][0] == 5
        out = run_select_query('select=z&where=float_str=eq.3.2')
        assert out[0][0] == 5
        out = run_select_query('select=z&where=float=gt.3.2')
        assert out[0][0] == 1
        # with quoting
        out = run_select_query("select=x&where=meh=eq.'.'")
        assert out[0][0] == 10
        out = run_select_query("select=x&where=lolly=eq.'()'")
        assert out[0][0] == 10
        out = run_select_query("select=x&where=wat=eq.'and:'")
        assert out[0][0] == 10
        out = run_select_query("select=x&where=meh2=eq.'()[],and:,or:. where=;'")
        assert out[0][0] == 107
        # ampersand
        out = run_select_query("select=x&where=being=eq.'arising&vanishing'")
        assert out[0][0] == 10
        # esacping single quotes
        out = run_select_query("select=x&where=loop=eq.'g\\'n kat oor die pad'")
        assert out[0][0] == 10
        # with a simple array
        out = run_select_query("select=z&where=b[0]=eq.1")
        assert out[0][0] == 5


        # ORDER
        if verbose:
            print('\n===> ORDER\n')
        # Note: postgres and sqlite treat NULLs different in ordering
        # postgres puts them first, sqlite puts them last, so be it
        # simple key
        out = run_select_query('select=x&where=x=not.is.null&order=x.desc')
        x_array = [[1900], [107], [88], [10]]
        assert out == x_array
        x_array.reverse()
        out = run_select_query('select=x&where=x=not.is.null&order=x.asc')
        assert out == x_array
        # array selections
        out = run_select_query('select=x,a&where=a.k1.r1[0]=not.is.null&order=a.k1.r1[0].desc')
        assert out[0][0] == 107
        out = run_select_query('select=x,a&where=a.k3[0|h]=not.is.null&order=a.k3[0|h].desc')
        assert out[0][0] == 107
        # timestamps
        out = run_select_query('select=x,timestamp&order=timestamp.desc')
        assert out[0][1] == '2020-10-14T20:20:34.388511'
        out = run_select_query('select=x,timestamp&order=timestamp.asc')
        assert out[0][1] == '2020-10-13T10:15:26.388573'

        # RANGE
        if verbose:
            print('\n===> RANGE\n')
        out = run_select_query('select=x&where=x=not.is.null&order=x.desc&range=0.2')
        assert out == [[1900], [107]]
        out = run_select_query('select=x&where=x=not.is.null&order=x.desc&range=1.2')
        assert out == [[107], [88]]

        # GROUP BY
        if verbose:
            print('\n===> GROUP BY\n')
        out = run_select_query('select=self,count(*)&group_by=self')
        assert len(out) == 2
        out = run_select_query('select=self,beneficial,count(*)&group_by=self,beneficial')
        assert len(out) == 4

        with pytest.raises(ParseError):
            run_select_query('select=self,x,count(*)&group_by=self,x&order=x.desc')

        with pytest.raises(ParseError):
            run_select_query('select=self,count(*)&group_by=self,x')

        # UPDATE
        if verbose:
            print('\n===> UPDATE\n')
        out = run_update_query('set=x&where=x=lt.1000', data={'x': 999})
        out = run_select_query('select=x&where=x=eq.999')
        assert out[0][0] == 999
        assert len(out) == 3
        out = run_update_query(
            'set=a&where=a.k1.r2=eq.90',
            data={'a': {'k1': {'r1': [33, 200], 'r2': 80 }}},
        )
        out = run_select_query('where=a.k1.r2=eq.80')
        assert len(out) == 1
        assert out[0]['a']['k1']['r2'] == 80
        with pytest.raises(ParseError):
            out = run_update_query('set=x&where=x=eq.1', data={})

        # multiple keys
        out = run_update_query(
            'set=x,y&where=float=eq.3.1',
            data={'x': 0, 'y': 1},
        )
        out = run_select_query('select=x,y&where=float=eq.3.1')
        assert len(out) == 1
        assert out[0][0] == 0
        assert out[0][1] == 1

        # setting to null
        out = run_update_query(
            'set=x&where=float=eq.3.1',
            data={'x': None},
        )
        out = run_select_query('select=x,y&where=float=eq.3.1')
        assert len(out) == 1
        assert out[0][0] == None

        # single quotes inside the payload
        out = run_update_query(
            "set=quotes_inside&where=wat=eq.'and:'",
            data={'quotes_inside': "this _has_ 'quotes'"},
        )
        out = run_select_query("select=quotes_inside&where=wat=eq.'and:'")
        assert len(out) == 1
        assert out[0][0] == 'this _has_ \'quotes\''

        # DELETE
        if verbose:
            print('\n===> DELETE\n')
        out = run_delete_query('where=x=lt.1000')
        assert out is True
        out = run_select_query('select=x&where=x=lt.1000')
        assert out == []
        out = run_delete_query('')
        with pytest.raises(Exception):
            out = run_delete_query('')

        # ALTER
        if verbose:
            print('\n===> ALTER\n')

        db.table_insert('some_table', data)

        # without an audit table
        out = run_alter_query("alter=name=eq.yet_another_table", "some_table")
        assert len(out["tables"]) == 1

        # with an audit table
        out = run_update_query(
            'set=x&where=float=eq.3.1',
            data={'x': None},
            table='yet_another_table'
        )
        out = run_alter_query("alter=name=eq.silly_table", "yet_another_table")
        assert len(out["tables"]) == 2

        # not permitted directly on an audit table
        with pytest.raises(OperationNotPermittedError):
            run_alter_query("alter=name=eq.new", audit_table("silly_table"))


    def test_sqlite(self):
        engine = sqlite_init('/tmp/api-test.db')
        self.run_backend_tests(
            self.data,
            engine,
            sqlite_session,
            SqliteQueryGenerator,
            SqliteBackend,
            self.verbose
        )

    def test_postgres(self) -> None:
        try:
            pool = postgres_init(
                {
                    "dbname": os.environ.get("PYSQURIL_POSTGRES_DB", "pysquril_db"),
                    "user": os.environ.get("PYSQURIL_POSTGRES_USER", "pysquril_user"),
                    "pw": os.environ.get("PYSQURIL_POSTGRES_PASSWORD", ""),
                    "host": os.environ.get("PYSQURIL_POSTGRES_HOST", "localhost"),
                }
            )
            pg_backend = PostgresBackend(pool)
            pg_backend.initialise()
            self.run_backend_tests(
                self.data,
                pool,
                postgres_session,
                PostgresQueryGenerator,
                PostgresBackend,
                self.verbose
            )
        except psycopg2.OperationalError:
            print("missing postgres db, run:")
            print("$ createuser pysquril_user")
            print("$ createdb -O pysquril_user pysquril_db")
            raise

class TestSqlBackend(unittest.TestCase):
    __test__ = False
    
    backend: Union[SqliteBackend, PostgresBackend]
    engine: Union[sqlite3.Connection, psycopg2.pool.SimpleConnectionPool]

    def test_audit(self) -> bool:
        test_table = "just_an_average_audit_test_table"

        pkey = "id"

        key_to_update = "key1"
        original_value = 5

        data = {pkey: 1, key_to_update: original_value, "key2": "a"}
        more_data = {pkey: 2, key_to_update: original_value, "key3": {"moar": "things"}}

        self.backend.table_insert(table_name=test_table, data=data)
        self.backend.table_insert(table_name=test_table, data=more_data)

        # update the table with the new data
        new_data = {key_to_update: original_value+1}
        message = "all the messages"

        self.backend.table_update(
            table_name=test_table,
            uri_query=f"set={key_to_update}&where={key_to_update}=eq.{original_value}&message={quote(message)}",
            data=new_data,
        )
        result = list(self.backend.table_select(table_name=test_table, uri_query=""))
        self.assertTrue(result)
        retrieved_data = result[0]
        self.assertEqual(retrieved_data, {**data, **new_data})
        self.assertNotEqual(retrieved_data[key_to_update], original_value)
        self.assertEqual(retrieved_data[key_to_update], new_data[key_to_update])

        # view update audit data
        result = list(self.backend.table_select(
            table_name=audit_table(test_table), uri_query="order=timestamp.asc")
        )
        self.assertTrue(result)
        audit_event = result[0]
        self.assertEqual(audit_event["previous"], data)
        self.assertEqual(audit_event["diff"], new_data)
        self.assertEqual(audit_event["event"], "update")
        self.assertTrue(audit_event["transaction_id"] is not None)
        self.assertTrue(audit_event["event_id"] is not None)
        self.assertTrue(audit_event["timestamp"] is not None)
        self.assertTrue(audit_event["query"] is not None)
        self.assertEqual(audit_event["message"], message)
        self.assertEqual(audit_event["identity"], TEST_REQUESTOR)
        self.assertEqual(audit_event["identity_name"], TEST_REQUESTOR_NAME)

        # restore updates
        with self.assertRaises(ParseError): # missing restore directive
            self.backend.table_restore(table_name=test_table, uri_query="")
        with self.assertRaises(ParseError): # missing primary key
            self.backend.table_restore(table_name=test_table, uri_query="restore")
        with self.assertRaises(ParseError): # still missing primary key
            self.backend.table_restore(table_name=test_table, uri_query="restore&primary_key=")

        # restore to a specific state, for a specific row
        message = "undoing mistakes"
        result = self.backend.table_restore(
            table_name=test_table,
            uri_query=f"restore&primary_key={pkey}&where=event_id=eq.{audit_event.get('event_id')}&message={quote(message)}"
        )
        self.assertEqual(len(result.get("updates")), 1)
        self.assertEqual(len(result.get("restores")), 0)
        result = list(self.backend.table_select(table_name=test_table, uri_query="where=id=eq.1"))
        self.assertEqual(result[0].get(key_to_update), original_value)
        result = list(self.backend.table_select(
            table_name=audit_table(test_table), uri_query="order=timestamp.desc")
        )
        self.assertEqual(result[0].get("message"), message)

        # delete a specific entry
        message = "bad data: must delete, & never repeat (tm)"
        self.backend.table_delete(
            table_name=test_table,
            uri_query=f"where=key3=not.is.null&message={quote(message)}"
        )
        result = list(self.backend.table_select(table_name=audit_table(test_table), uri_query=""))
        self.assertEqual(len(result), 4)
        result = list(self.backend.table_select(
            table_name=audit_table(test_table), uri_query="order=timestamp.desc")
        )
        self.assertEqual(result[0].get("message"), message)

        # restore the deleted entry
        result = self.backend.table_restore(
            table_name=test_table,
            uri_query=f"restore&primary_key={pkey}&where=event=eq.delete"
        )
        self.assertEqual(len(result.get("updates")), 0)
        self.assertEqual(len(result.get("restores")), 1)
        result = list(self.backend.table_select(table_name=test_table, uri_query="where=id=eq.2"))
        self.assertTrue(result[0] is not None)

        # delete the table
        self.backend.table_delete(table_name=test_table, uri_query="")

        # check that the deletes are in the audit
        result = list(self.backend.table_select(table_name=audit_table(test_table), uri_query=""))
        self.assertEqual(len(result), 7)

        # restore everything
        result = self.backend.table_restore(table_name=test_table, uri_query=f"restore&primary_key={pkey}")
        self.assertTrue(result is not None)
        result = list(self.backend.table_select(table_name=test_table, uri_query="order=id.asc"))
        self.assertEqual(result[0], data)
        self.assertEqual(result[1], more_data)

        # delete the table (again)
        self.backend.table_delete(table_name=test_table, uri_query="")

        # try to retrieve deleted table
        select = self.backend.table_select(table_name=test_table, uri_query="")
        with self.assertRaises((sqlite3.OperationalError, psycopg2.errors.UndefinedTable)):
            next(select)

        # delete the audit table
        self.backend.table_delete(table_name=audit_table(test_table), uri_query="")
        
        # try to retrieve deleted table's audit table
        select = self.backend.table_select(table_name=audit_table(test_table), uri_query="")
        with self.assertRaises((sqlite3.OperationalError, psycopg2.errors.UndefinedTable)):
            next(select)

        # test deleting an entire table, without any updates
        # automatic audit table creation on delete
        # use a nested primary key, to test restores with such keys
        some_data = {"pk": {"id": 0}, "lol": None, "cat": [1, 2]}
        some_more_data = {"pk": {"id": 1}, "neither-lol-not-not-lol": None, "cat": []}
        some_table = "yay"
        self.backend.table_insert(table_name=some_table, data=some_data)
        self.backend.table_insert(table_name=some_table, data=some_more_data)
        self.backend.table_update(
            table_name=some_table,
            uri_query=f"set=lol&where=pk.id=eq.0",
            data={"lol": "wat"},
        )
        self.backend.table_delete(table_name=some_table, uri_query="")
        audit = list(self.backend.table_select(table_name=audit_table(some_table), uri_query=""))
        self.assertEqual(len(audit), 3)
        nested_result = self.backend.table_restore(
            table_name=some_table, uri_query=f"restore&primary_key=pk.id"
        )
        self.assertEqual(len(nested_result.get("restores")), 2)
        self.assertEqual(len(nested_result.get("updates")), 0)
        self.backend.table_delete(table_name=some_table, uri_query="")
        self.backend.table_delete(table_name=audit_table(some_table), uri_query="")


        # test backup retention enforcement
        backup_table = "backedup"
        self.backend.backup_days = 1
        self.backend.table_insert(table_name=backup_table, data={"breathe-in": "long", "id": 0})
        self.backend.table_insert(table_name=backup_table, data={"breathe-out": "long", "id": 1})
        self.backend.table_delete(table_name=backup_table, uri_query="")

        # within the retention period
        audit = list(self.backend.table_select(table_name=audit_table(backup_table), uri_query=""))
        self.assertEqual(len(audit), 2)
        result = self.backend.table_restore(
            table_name=backup_table, uri_query=f"restore&primary_key=id",
        )
        self.assertEqual(len(result.get("restores")), 2)
        original = list(self.backend.table_select(table_name=backup_table, uri_query=""))
        self.assertTrue(len(original), 2)

        # cleanup
        self.backend.table_delete(table_name=backup_table, uri_query="")
        self.backend.table_delete(table_name=audit_table(backup_table), uri_query="")


        # outside the retention period
        not_backup_table = "notbackedup"
        self.backend.backup_days = 1
        self.backend.table_insert(table_name=not_backup_table, data={"breathe-in": "short", "id": 0})
        self.backend.table_insert(table_name=not_backup_table, data={"breathe-out": "short", "id": 1})
        self.backend.table_delete(table_name=not_backup_table, uri_query="")

        # now adjust the audit timestamps to fall outside the retention period
        target = (datetime.datetime.now() - timedelta(days=2)).isoformat()
        if isinstance(self.backend, SqliteBackend):
            new = json.dumps({"timestamp": target})
            update_query = f"update {audit_table(not_backup_table)} set data = json_patch(data, '{new}')"
        elif isinstance(self.backend, PostgresBackend):
            update_query = f"update {audit_table(not_backup_table)} set data = jsonb_set(data, '{{timestamp}}', '\"{target}\"')"
        with self.session_func(self.engine) as session:
            session.execute(update_query)

        # should not be able to view audit or restore data
        audit = list(self.backend.table_select(table_name=audit_table(not_backup_table), uri_query=""))
        self.assertEqual(len(audit), 0)
        result = self.backend.table_restore(
            table_name=not_backup_table, uri_query=f"restore&primary_key=id",
        )
        self.assertEqual(len(result.get("restores")), 0)

        # cleanup
        self.backend.table_delete(table_name=audit_table(not_backup_table), uri_query="")

        # delete without audit
        table_without_audit = "without_audit"
        self.backend.table_insert(table_name=table_without_audit, data={"breathe": "calming", "id": 0})
        self.backend.table_delete(table_name=table_without_audit, uri_query="", audit=False)
        with pytest.raises((psycopg2.errors.UndefinedTable, sqlite3.OperationalError)):
            audit = list(
                self.backend.table_select(
                    table_name=audit_table(table_without_audit),
                    uri_query="",
                )
            )

        # audit for create and read
        verbose_table = "table_with_full_audit"
        try:
            self.backend.table_delete(table_name=verbose_table, uri_query="")
            self.backend.table_delete(table_name=audit_table(verbose_table), uri_query="")
        except Exception:
            pass
        self.backend.table_insert(
            table_name=verbose_table,
            data={"observing": "mind objects", "id": 0},
            audit=True,
        )
        self.backend.table_select(
            table_name=verbose_table, uri_query="select=observing", audit=True,
        )
        audit = list(
            self.backend.table_select(
                table_name=audit_table(verbose_table),
                uri_query="",
            )
        )
        self.assertEqual(len(audit), 2)
        self.backend.table_delete(table_name=audit_table(verbose_table), uri_query="")

    def test_all_view(self) -> bool:
        tenant1 = "p11"
        tenant2 = "p12"
        tenant3 = "p13"
        table_name = "A"
        for tenant in [tenant1, tenant2, tenant3]:
            view_backend = self.backend_class(
                self.engine, schema=tenant, schema_pattern="p"
            )
            try:
                view_backend.table_delete(table_name=table_name, uri_query="")
            except Exception as e:
                pass
            view_backend.table_insert(
                table_name,
                data={"id": str(uuid.uuid4()), "data": f"yes"},
                update_all_view=True,
            )
        all_backend = self.backend_class(
            self.engine, schema="all", schema_pattern="p"
        )
        result = list(all_backend.table_select(table_name, ""))
        self.assertEqual(len(result), 3)

class TestSqliteBackend(TestSqlBackend):
    __test__ = True

    def setUp(self) -> None:
        self.directory = tempfile.gettempdir()
        self.file = f"{__package__}_test.db"
        self.engine = sqlite_init(f"{self.directory}/{self.file}")
        self.backend = SqliteBackend(
            self.engine, requestor=TEST_REQUESTOR, requestor_name=TEST_REQUESTOR_NAME
        )
        self.session_func = sqlite_session
        self.backend_class = SqliteBackend
    
    def tearDown(self) -> None:
        self.engine.close()
        if os.path.exists(f"{self.directory}/{self.file}"):
            os.remove(f"{self.directory}/{self.file}")

class TestPostgresBackend(TestSqlBackend):
    __test__ = True

    def setUp(self) -> None:
        self.engine = postgres_init(
            {
                "dbname": os.environ.get("PYSQURIL_POSTGRES_DB", "pysquril_db"),
                "user": os.environ.get("PYSQURIL_POSTGRES_USER", "pysquril_user"),
                "pw": os.environ.get("PYSQURIL_POSTGRES_PASSWORD", ""),
                "host": os.environ.get("PYSQURIL_POSTGRES_HOST", "localhost"),
            }
        )
        self.backend = PostgresBackend(
            self.engine, requestor=TEST_REQUESTOR, requestor_name=TEST_REQUESTOR_NAME
        )
        self.backend.initialise()
        self.session_func = postgres_session
        self.backend_class = PostgresBackend

    def tearDown(self) -> None:
        self.engine.closeall()




"""SQURIL - Structured Query URI Language."""

import json
import re

from abc import ABC, abstractmethod
from typing import Optional, Union, Callable, Any
from urllib.parse import unquote

from pysquril.exc import ParseError


def previous_element(in_list: list, current_idx: int) -> Any:
    try:
        element = in_list[current_idx - 1]
    except IndexError:
        element = ""
    return element

class SelectElement(ABC):
    @property
    @abstractmethod
    def name(self) -> str:
        pass
    @property
    @abstractmethod
    def regex(self) -> str:
        pass


class BaseSelectElement(SelectElement):

    name = None
    regex = None

    def __init__(self, element: str) -> None:
        self.element = element
        self.bare_key = self.create_bare_key(self.element)
        self.sub_selections = self.create_sub_selections(self.element)
        self.idx = self.create_idx(self.element)

    def create_bare_key(self, element: str) -> Optional[list]:
        return element.split('[')[0] if '[' in element else None

    def create_sub_selections(self, element: str) -> list:
        return element.split('|')[1].replace(']', '').split(',') if '|' in element else []

    def create_idx(self, element: str) -> Optional[str]:
        if '[' in element and '|' in element:
            return re.sub(r'.+\[(.*)\|(.*)\]', r'\1', element)
        elif '[' in element and '|' not in element:
            return re.sub(r'.+\[(.*)\]', r'\1', element)
        else:
            return None


class Key(BaseSelectElement):
    name = 'key'
    regex = r'[^\[\]]+$'


class ArraySpecific(BaseSelectElement):
    name = 'array.specific'
    regex = r'.+\[[0-9]+\]$'


class ArraySpecificSingle(BaseSelectElement):
    name = 'array.specific.single'
    regex = r'.+\[[0-9]+\|[^,]+\]$'


class ArraySpecificMultiple(BaseSelectElement):
    name = 'array.specific.multiple'
    regex = r'.+\[[0-9]+\|.+,.+\]$'


class ArrayBroadcastSingle(BaseSelectElement):
    name = 'array.broadcast.single'
    regex = r'.+\[\*\|[^,]+\]$'


class ArrayBroadcastMultiple(BaseSelectElement):
    name = 'array.broadcast.multiple'
    regex = r'.+\[\*\|.+,.+\]$'


class SelectTerm(object):

    supported_functions = [
        'count', 'avg', 'sum', 'min', 'max', 'min_ts', 'max_ts',
    ]

    element_classes = [
        Key,
        ArraySpecific,
        ArraySpecificSingle,
        ArraySpecificMultiple,
        ArrayBroadcastSingle,
        ArrayBroadcastMultiple
    ]

    def __init__(self, original: str) -> None:
        self.func, self.original = self.strip_function(original)
        self.bare_term = self.original.split('[')[0]
        self.parsed = self.parse_elements()

    def strip_function(self, term: str) -> tuple:
        func = None
        for sf in self.supported_functions:
            if term.startswith(f"{sf}("):
                func = sf
                term = term.replace(f"{sf}(", "")[:-1]
                break
        return func, term

    def parse_elements(self) -> list:
        out = []
        parts = self.original.split('.')
        for element in parts:
            element_instance = None
            found = False
            for ElementClass in self.element_classes:
                if re.match(ElementClass.regex, element):
                    if found:
                        msg = f'Could not uniquely identify {element} - already matched with {found}'
                        raise ParseError(msg)
                    element_instance = ElementClass(element)
                    found = ElementClass.name
            if not element_instance:
                raise ParseError(f'Could not parse {element}')
            out.append(element_instance)
        return out


class GroupByTerm(SelectTerm):

    element_classes = [
        Key,
        ArraySpecific,
        ArraySpecificSingle,
    ]

    def strip_function(self, term: str) -> tuple:
        func = None
        for sf in self.supported_functions:
            if term.startswith(f"{sf}("):
                raise ParseError("group_by keys cannot contain fuctions")
        return func, term


class WhereElement(object):

    def __init__(
        self,
        groups: list,
        combinator: str,
        term: str,
        op: str,
        val: str,
    ) -> None:
        self.groups_start, self.groups_end = self.categorise_groups(groups)
        self.combinator = combinator
        self.select_term = SelectTerm(term)
        self.op = op
        self.val = val

    def categorise_groups(self, groups: list) -> tuple:
        start, end = [], []
        for bracket in groups:
            if bracket == '(':
                start.append(bracket)
            elif bracket == ')':
                end.append(bracket)
        return start, end


class WhereTerm(object):

    combinators = ['and:', 'or:']

    def __init__(self, original: str) -> None:
        self.original = original
        self.parsed = self.parse_elements_quoted()

    def parse_elements_quoted(self) -> list:
        element = self.original
        temp = ""
        groups = []
        # find groups, remove brackets
        is_quoted = False
        for idx, token in enumerate(element):
            previous = previous_element(element, idx)
            if token == "'" and previous != "\\":
                is_quoted = not is_quoted
            if token in ['(', ')'] and not is_quoted:
                groups.append(token)
            else:
                temp += token
        element = temp # with groups removed
        # find and remove logical operators
        combinator = None
        for c in self.combinators:
            if element.startswith(c):
                combinator = c.replace(':', '')
                element = element[len(c):]
        # find term, operator, and value: {term}={op}.{val}
        term, term_found = "", False
        op, op_found = "", False
        val = ""
        negated_ops = 0
        for idx, token in enumerate(element):
            previous = previous_element(element, idx)
            if not term_found:
                if token != "=":
                    term += token
                else:
                    term_found = True
            else:
                if not op_found:
                    if token != ".":
                        op += token
                    else:
                        if op.startswith("not"):
                            if token == ".":
                                if negated_ops == 1:
                                    op_found = True
                                    continue
                                negated_ops += 1
                                op += token
                        else:
                            op_found = True
                else:
                    if (token == "'" and previous != "\\") or token == "\\":
                        continue
                    elif token == "'" and previous == "\\":
                        val += "''"
                    else:
                        val += token
        if op == "not.is":
            op = "is.not"
        return [WhereElement(groups, combinator, term, op, val)]


class OrderElement(object):

    def __init__(self, term: str, direction: str) -> None:
        self.select_term = SelectTerm(term)
        self.direction = direction


class OrderTerm(object):

    def __init__(self, original: str) -> None:
        self.original = original
        self.parsed = self.parse_elements()

    def parse_elements(self) -> list:
        parts = self.original.split('.')
        term = '.'.join(parts[:-1])
        direction = parts[-1]
        return [OrderElement(term, direction)]


class RangeElement(object):

    def __init__(self, start: str, end: str) -> None:
        self.start = start
        self.end = end


class RangeTerm(object):

    def __init__(self, original: str) -> None:
        self.original = original
        self.parsed = self.parse_elements()

    def parse_elements(self) -> list:
        start, end = self.original.split('.')
        return [RangeElement(start, end)]


class SetElement(object):

    def __init__(self, term: str) -> None:
        self.select_term = SelectTerm(term)
        if not isinstance(self.select_term.parsed[0], Key):
            raise ParseError(f'{term} must be an instance of Key')
        if not len(self.select_term.parsed) == 1:
            # note: relaxing this would require changes to table_restore
            raise ParseError(f'SetElements can only be top level keys - {term} is nested')


class SetTerm(object):

    def __init__(self, original: str) -> None:
        self.original = original
        self.parsed = self.parse_elements()

    def parse_elements(self) -> list:
        return [SetElement(self.original)]


class Clause(object):

    term_class = None

    def __init__(self, original: str) -> None:
        self.original = original
        self.parsed = self.parse_terms()
        self._enforce_constraints()

    def split_clause(self) -> list:
        braces_open = False
        temp = ""
        parts = []
        is_quoted = False
        previous = ""
        for idx, token in enumerate(self.original):
            previous = previous_element(self.original, idx)
            if token == "'" and previous != "\\":
                is_quoted = not is_quoted
            if not is_quoted:
                if token == '[':
                    braces_open = True
                if token == ']':
                    braces_open = False
            if token == ",":
                if braces_open or is_quoted:
                    temp += token
                else:
                    parts.append(temp)
                    temp = ""
            else:
                temp += token
        if temp:
            parts.append(temp)
        return parts

    def parse_terms(self) -> list:
        out = []
        terms = self.split_clause()
        for term in terms:
            out.append(self.term_class(term))
        return out

    def _enforce_constraints(self) -> None:
        """
        Implemented if the specific clause must enforce
        contraints not shared by all.

        """
        pass


class SelectClause(Clause):
    term_class = SelectTerm

class WhereClause(Clause):
    term_class = WhereTerm

class OrderClause(Clause):
    term_class = OrderTerm

class RangeClause(Clause):
    term_class = RangeTerm

class SetClause(Clause):
    term_class = SetTerm

class GroupByClause(Clause):
    term_class = SelectTerm

class AlterClause(Clause):
    term_class = WhereTerm

    def _enforce_constraints(self) -> None:
        """
        Only rename is supported thus far.

        """
        term = self.parsed[0]
        element = term.parsed[0]
        if element.select_term.bare_term != "name":
            raise ParseError("alter statements limited to `name` attribute")
        if element.op != "eq":
            raise ParseError(f"rename requires `eq` operator, not {element.op}")


class Message(object):

    def __init__(self, original: str) -> None:
        self.original = original
        self.parsed = self.parse_message(original)

    def parse_message(self, message: str) -> str:
        out = ""
        if unquote(message) != message:
            out = unquote(message)
        else:
            for idx, token in enumerate(message):
                previous = previous_element(message, idx)
                if (token == "'" and previous != "\\") or token == "\\":
                    continue
                elif token == "'" and previous == "\\":
                    out += "''"
                else:
                    out += token
        return out


class UriQuery(object):

    """
    Lex and parse a URI query into a UriQuery object:

        Query
            -> Clause(s)
                -> [Term(s)]
                    -> [Element(s)]

    """

    def __init__(
        self,
        table_name: str,
        uri_query: str,
        data: Union[list, dict] = None,
    ) -> None:
        self.table_name = table_name
        self.original = uri_query
        self.data = data
        self.select = self.parse_clause(prefix='select=', Cls=SelectClause)
        self.where = self.parse_clause(prefix='where=', Cls=WhereClause)
        self.order = self.parse_clause(prefix='order=', Cls=OrderClause)
        self.range = self.parse_clause(prefix='range=', Cls=RangeClause)
        self.set = self.parse_clause(prefix='set=', Cls=SetClause)
        self.alter = self.parse_clause(prefix='alter=', Cls=AlterClause)
        self.group_by = self.parse_clause(prefix='group_by=', Cls=GroupByClause)
        self.message = self.parse_message()
        if self.group_by:
            if self.order:
                raise ParseError("ordering not supported for group_by")
            group_by_keys = self.group_by.split_clause()
            select_keys = self.select.split_clause()
            if not set(group_by_keys).intersection(select_keys) == set(group_by_keys):
                raise ParseError("group by keys must be used in select")

    def _slice(
        self,
        *,
        target: str,
        positions: list,
    ) -> list:
        """
        Slice a string, at given index values, with
        a given character, e.g.:

        _slice(
            target="abc&123&890",
            positions=[4, 8],
        ) -> ["abc", "123", "890"]

        """
        out = []
        part = ""
        last = False
        if not positions:
            return [target]
        position = positions.pop(0)
        for idx, char in enumerate(target):
            if idx < position:
                part += char
            elif idx == position:
                out.append(part)
                part = ""
                try:
                    position = positions.pop(0)
                except IndexError:
                    last = True
                continue
            elif last:
                part = target[position + 1:]
                out.append(part)
                break
        return out

    def _index_clauses(self, uri_query: str) -> list:
        """
        Find the index positions in the uri_query
        which mark the separator `&` between clauses,
        taking into account that ampersand can be
        used inside quoted values of where clauses.

        """
        positions = []
        is_quoted = False
        for idx, token in enumerate(uri_query):
            if token == "'":
                is_quoted = not is_quoted
            if token == "&" and not is_quoted:
                positions.append(idx)
        return positions

    def parse_clause(self, *, prefix: str, Cls: Clause) -> Clause:
        parts = self._slice(
            target=self.original, positions=self._index_clauses(self.original)
        )
        for part in parts:
            if part.startswith(prefix):
                return Cls(part[len(prefix):])

    def parse_message(self) -> str:
        prefix = "message="
        message = None
        parts = self._slice(
            target=self.original, positions=self._index_clauses(self.original)
        )
        for part in parts:
            if part.startswith(prefix):
                message = Message(part[len(prefix):]).parsed
        return message